# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                          '
#                                                                                        '
#  This file is part of octree-creation-app package.                                     '
#                                                                                        '
#  octree-creation-app is distributed under the terms and conditions of the MIT License  '
#  (see LICENSE file at the root of this source code package).                           '
#                                                                                        '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
from __future__ import annotations

import string
import warnings
from pathlib import Path
from typing import Any, ClassVar

import numpy as np
from geoapps_utils.driver.data import BaseData
from geoh5py.groups import UIJsonGroup
from geoh5py.objects import Points
from geoh5py.ui_json import InputFile
from pydantic import (
    BaseModel,
    ConfigDict,
    field_serializer,
    field_validator,
    model_serializer,
    model_validator,
)
from typing_extensions import Self

import octree_creation_app
from octree_creation_app import assets_path


defaults_ifile = InputFile.read_ui_json(
    assets_path() / "uijson/octree_mesh.ui.json", validate=False
)
default_ui_json = defaults_ifile.ui_json
defaults = defaults_ifile.data


class OctreeParams(BaseData):
    """
    Octree creation parameters.

    :param objects: Object used to define the core of the mesh.
    :param depth_core: Limit the depth of the core of the mesh.
    :param diagonal_balance: Whether to limit the cell size change
        to one level in the transition between diagonally adjacent
        cells.
    :param minimum_level: Provides a minimum level of refinement for
        the whole mesh to prevent excessive coarsenin in padding
        regions.
    :param u_cell_size: Cell size in the x-direction.
    :param v_cell_size: Cell size in the y-direction.
    :param w_cell_size: Cell size in the z-direction.
    :param horizontal_padding: Padding in the x and y directions.
    :param vertical_padding: Padding in the z direction.
    :param refinements: List of refinements to be applied.
    """

    model_config = ConfigDict(
        frozen=True,
        arbitrary_types_allowed=True,
    )

    name: ClassVar[str] = "Octree_Mesh"
    default_ui_json: ClassVar[Path] = assets_path() / "uijson/octree_mesh.ui.json"

    version: str = octree_creation_app.__version__
    title: str = "Octree Mesh Creator"
    run_command: str = "octree_creation_app.driver"
    conda_environment: str = "octree_creation_app"
    objects: Points
    depth_core: float = 500.0
    ga_group_name: str = "Octree Mesh"  # TODO: Alias this in uijson (mesh_name)
    diagonal_balance: bool = True
    minimum_level: int = 8
    u_cell_size: float = 25.0
    v_cell_size: float = 25.0
    w_cell_size: float = 25.0
    horizontal_padding: float = 500.0
    vertical_padding: float = 200.0
    refinements: list[RefinementParams | None] | None = None
    out_group: UIJsonGroup | None = None

    @model_validator(mode="before")
    @classmethod
    def collect_refinements(cls, values: dict):
        """Collect refinements from the input dictionary."""
        if "refinements" not in values:
            refinements = collect_refinements_from_dict(values)
            if refinements:
                msg = (
                    "Detected deprecated 'Refinement A property' style refinements,"
                    " converting to a list of dictionaries."
                )
                warnings.warn(msg)
            values["refinements"] = refinements

        return values

    @model_serializer(mode="wrap")
    def distribute_refinements(self, handler, info):
        """Convert refinements to a individual parameters."""
        dump = handler(self, info)
        refinements = dump.pop("refinements")
        refinement_params: dict[str, Any] = {}
        for i, group in enumerate(refinements):
            group_id = string.ascii_uppercase[i]
            if group is None:
                refinement_params[f"Refinement {group_id} object"] = None
                refinement_params[f"Refinement {group_id} levels"] = None
                refinement_params[f"Refinement {group_id} horizon"] = None
                refinement_params[f"Refinement {group_id} distance"] = None
            else:
                for param, value in group.items():
                    param_type = "object" if param == "refinement_object" else param
                    param_name = f"Refinement {group_id} {param_type}"
                    refinement_params[param_name] = value

        return dict(dump, **refinement_params)

    @classmethod
    def build(cls, input_data: InputFile | dict) -> Self:
        """
        Build a dataclass from a dictionary or InputFile.

        :param input_data: Dictionary of parameters and values.

        :return: Dataclass of application parameters.
        """

        data = input_data

        if isinstance(input_data, InputFile) and input_data.data is not None:
            data = input_data.data.copy()

        if not isinstance(data, dict):
            raise TypeError("Input data must be a dictionary or InputFile.")

        kwargs = OctreeParams.collect_input_from_dict(cls, data)  # type: ignore
        out = cls(**kwargs)
        if isinstance(input_data, InputFile):
            out._input_file = input_data

        return out

    @classmethod
    def collect_input_from_dict(cls, base_model: type[BaseModel], data: dict[str, Any]):
        """
        Recursively replace BaseModel objects with dictionary of 'data' values.

        :param base_model: BaseModel object holding data and possibly other nested
            BaseModel objects.
        :param data: Dictionary of parameters and values without nesting structure.
        """

        update = super().collect_input_from_dict(base_model, data)
        update["refinements"] = collect_refinements_from_dict(data)

        return update

    def get_padding(self) -> list:
        """
        Utility to get the padding values as a list of padding along each axis.
        """
        return [
            [
                self.horizontal_padding,
                self.horizontal_padding,
            ],
            [
                self.horizontal_padding,
                self.horizontal_padding,
            ],
            [self.vertical_padding, self.vertical_padding],
        ]


class RefinementParams(BaseModel):
    """
    Refinement parameters.

    :param object: Object used to define the core of the mesh.
    :param levels: Levels of refinement.
    :param horizon: Whether the refinement is a surface or radial.
    :param distance: Distance from the object to refine.
    """

    model_config = ConfigDict(
        frozen=True,
        arbitrary_types_allowed=True,
    )
    refinement_object: Points
    levels: list[int] = [4, 2]
    horizon: bool = False
    distance: float | None = np.inf

    @field_validator("levels", mode="before")
    @classmethod
    def int_2_list(cls, levels: int | list[int]):
        if isinstance(levels, int):
            levels = [levels]
        return levels

    @field_validator("levels", mode="before")
    @classmethod
    def string_2_list(cls, levels: str | list[int]):
        if isinstance(levels, str):
            levels = [int(level) for level in levels.split(",")]
        return levels

    @field_serializer("levels")
    def list_to_string(self, value):
        return ", ".join(str(v) for v in value)


def collect_refinements_from_dict(data: dict) -> list[dict | None]:
    """Collect active refinement dictionaries from input dictionary."""
    refinements: list[dict | None] = []
    for identifier in refinement_identifiers(data):
        refinement_params = {}
        for param in ["object", "levels", "horizon", "distance"]:
            name = f"refinement_{param}" if param == "object" else param
            refinement_name = f"Refinement {identifier} {param}"
            refinement_params[name] = data.get(refinement_name, None)

        if refinement_params["refinement_object"] is None:
            refinements.append(None)
        else:
            refinements.append(refinement_params)

    return refinements


def refinement_identifiers(data: dict) -> list[str]:
    """Return identifiers for active refinements (object not none)."""
    refinements = [k for k in data if "Refinement" in k]
    active = [k for k in refinements if "object" in k]
    return np.unique([k.split(" ")[1] for k in active])



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                          '
#                                                                                        '
#  This file is part of octree-creation-app package.                                     '
#                                                                                        '
#  octree-creation-app is distributed under the terms and conditions of the MIT License  '
#  (see LICENSE file at the root of this source code package).                           '
#                                                                                        '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
from __future__ import annotations


REFINEMENT_KEY = "refinement"

template_dict: dict[str, dict] = {
    "object": {
        "groupOptional": True,
        "enabled": False,
        "group": "Refinement A",
        "label": "Object",
        "meshType": [
            "{202c5db1-a56d-4004-9cad-baafd8899406}",
            "{6a057fdc-b355-11e3-95be-fd84a7ffcb88}",
            "{f26feba3-aded-494b-b9e9-b2bbcbe298e1}",
            "{b99bd6e5-4fe1-45a5-bd2f-75fc31f91b38}",
            "{0b639533-f35b-44d8-92a8-f70ecff3fd26}",
            "{9b08bb5a-300c-48fe-9007-d206f971ea92}",
            "{19730589-fd28-4649-9de0-ad47249d9aba}",
            "{b3a47539-0301-4b27-922e-1dde9d882c60}",
            "{a81c6b0a-f290-4bc8-b72d-60e59964bfe8}",
            "{41018a45-01a0-4c61-a7cb-9f32d8159df4}",
            "{deebe11a-b57b-4a03-99d6-8f27b25eb2a8}",
            "{275ecee9-9c24-4378-bf94-65f3c5fbe163}",
        ],
        "value": None,
        "tooltip": "Object used to refine the mesh. Refinement strategy varies "
        "depending on the object type. See documentation for details.",
    },
    "levels": {
        "enabled": True,
        "group": "Refinement A",
        "label": "Levels",
        "value": "4, 4, 4",
        "tooltip": "Number of consecutive cells requested at each octree level. "
        "See documentation for details.",
    },
    "horizon": {
        "enabled": True,
        "group": "Refinement A",
        "label": "Use as horizon",
        "tooltip": "Object vertices are triangulated. Refinement levels are "
        "applied as depth layers.",
        "value": False,
    },
    "distance": {
        "enabled": False,
        "group": "Refinement A",
        "dependency": "horizon",
        "dependencyType": "enabled",
        "label": "Distance",
        "tooltip": "Radial horizontal distance to extend the refinement "
        "around each vertex.",
        "value": 1000.0,
    },
}



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                          '
#                                                                                        '
#  This file is part of octree-creation-app package.                                     '
#                                                                                        '
#  octree-creation-app is distributed under the terms and conditions of the MIT License  '
#  (see LICENSE file at the root of this source code package).                           '
#                                                                                        '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
from __future__ import annotations

from pathlib import Path

from geoapps_utils.utils.importing import assets_path as assets_path_impl


__version__ = "0.3.0-alpha.1"


def assets_path() -> Path:
    """Return the path to the assets folder."""

    return assets_path_impl(__file__)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                          '
#                                                                                        '
#  This file is part of octree-creation-app package.                                     '
#                                                                                        '
#  octree-creation-app is distributed under the terms and conditions of the MIT License  '
#  (see LICENSE file at the root of this source code package).                           '
#                                                                                        '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import logging
import sys

import numpy as np
from discretize import TreeMesh
from discretize.utils import mesh_builder_xyz
from geoapps_utils.driver.driver import BaseDriver
from geoapps_utils.utils.locations import get_locations
from geoh5py.objects import Curve, ObjectBase, Octree, Points, Surface
from geoh5py.objects.surveys.direct_current import BaseElectrode
from geoh5py.shared.utils import fetch_active_workspace
from scipy import interpolate
from scipy.spatial import Delaunay, QhullError, cKDTree

from octree_creation_app.params import OctreeParams, RefinementParams
from octree_creation_app.utils import densify_curve, surface_strip, treemesh_2_octree


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class OctreeDriver(BaseDriver):
    """Driver for octree mesh creation."""

    _params_class = OctreeParams
    _validations: dict = {}

    def __init__(self, params: OctreeParams):
        super().__init__(params)
        self.params: OctreeParams = params

    def run(self) -> Octree:
        """Create an octree mesh from input values."""
        with fetch_active_workspace(self.params.geoh5, mode="r+"):
            logger.info("Creating octree mesh from params . . .")
            octree = self.octree_from_params(self.params)
            self.update_monitoring_directory(octree)
            logger.info("Done.")

        return octree

    @staticmethod
    def octree_from_params(params: OctreeParams) -> Octree:
        """Create an Octree object from input parameters."""
        treemesh = OctreeDriver.treemesh_from_params(params)
        octree = treemesh_2_octree(
            params.geoh5, treemesh, name=params.ga_group_name, parent=params.out_group
        )
        return octree

    @staticmethod
    def treemesh_from_params(params: OctreeParams) -> TreeMesh:
        """Create a TreeMesh object from input parameters."""
        logger.info("Setting the mesh extent . . .")
        mesh = OctreeDriver.base_treemesh(params)

        logger.info("Applying minimum level refinement . . .")
        mesh = OctreeDriver.refine_minimum_level(mesh, params.minimum_level)

        logger.info("Applying extra refinements . . .")
        if params.refinements is not None:
            OctreeDriver.refine_objects(mesh, params.refinements)

        logger.info("Finalizing . . .")
        mesh.finalize()

        return mesh

    @staticmethod
    def base_treemesh(params: OctreeParams) -> TreeMesh:
        """Create a base TreeMesh object from extents."""

        entity = params.objects
        if hasattr(entity, "complement"):
            vertices = np.vstack([entity.vertices, entity.complement.vertices])
        else:
            vertices = entity.vertices

        mesh: TreeMesh = mesh_builder_xyz(
            vertices,
            [
                params.u_cell_size,
                params.v_cell_size,
                params.w_cell_size,
            ],
            padding_distance=params.get_padding(),
            mesh_type="tree",
            depth_core=params.depth_core,
            tree_diagonal_balance=params.diagonal_balance,
        )

        deltas = OctreeDriver.tree_offset(mesh, vertices)
        mesh.origin += deltas
        return mesh

    @staticmethod
    def refine_minimum_level(mesh: TreeMesh, minimum_level: int) -> TreeMesh:
        """Refine a TreeMesh with the minimum level of refinement."""
        minimum_level = OctreeDriver.minimum_level(mesh, minimum_level)
        mesh.refine(minimum_level, finalize=False)
        return mesh

    @staticmethod
    def refine_objects(
        mesh: TreeMesh, refinements: list[RefinementParams | None]
    ) -> TreeMesh:
        """
        Refine by object or object + complement.

        :param mesh: Tree mesh to refine.
        :param refinements: List of refinements to apply.
        """
        for refinement in refinements:
            if refinement is None:
                continue
            kwargs = refinement.model_dump()
            kwargs["levels"] = [int(k) for k in kwargs["levels"].split(",")]
            refinement_object = [kwargs.pop("refinement_object")]
            if hasattr(refinement_object[0], "complement"):
                refinement_object.append(refinement_object[0].complement)

            for obj in refinement_object:
                mesh = OctreeDriver.refine_by_object_type(
                    mesh=mesh,
                    refinement_object=obj,
                    **kwargs,
                )

        return mesh

    @staticmethod
    def minimum_level(mesh: TreeMesh, level: int) -> int:
        """Computes the minimum level of refinement for a given tree mesh."""
        return max([1, mesh.max_level - level + 1])

    @staticmethod
    def refine_by_object_type(
        mesh: TreeMesh,
        refinement_object: ObjectBase,
        levels: list[int],
        *,
        horizon: bool,
        distance: float | None,
    ) -> TreeMesh:
        """Refine Treemesh as a based on object type."""
        if horizon:
            try:
                mesh = OctreeDriver.refine_tree_from_surface(
                    mesh,
                    refinement_object,
                    levels,
                    max_distance=np.inf if distance is None else distance,
                )
            except QhullError:
                base_cell_size = np.min([h.min() for h in mesh.h])
                mesh = OctreeDriver.refine_tree_from_surface(
                    mesh,
                    surface_strip(refinement_object, 2 * base_cell_size),
                    levels,
                    max_distance=np.inf if distance is None else distance,
                )

        elif isinstance(refinement_object, Curve):
            mesh = OctreeDriver.refine_tree_from_curve(mesh, refinement_object, levels)

        elif isinstance(refinement_object, Surface):
            mesh = OctreeDriver.refine_tree_from_triangulation(
                mesh, refinement_object, levels
            )

        elif isinstance(refinement_object, Points):
            mesh = OctreeDriver.refine_tree_from_points(
                mesh,
                refinement_object,
                levels,
            )

        else:
            raise NotImplementedError(
                f"Refinement for object {type(refinement_object)} is not implemented."
            )

        return mesh

    @staticmethod
    def refine_tree_from_curve(
        mesh: TreeMesh,
        curve: Curve,
        levels: list[int] | np.ndarray,
        *,
        finalize: bool = False,
    ) -> TreeMesh:
        """
        Refine a tree mesh along the segments of a curve densified by the
        mesh cell size.

        :param mesh: Tree mesh to refine.
        :param curve: Curve object to use for refinement.
        :param levels: Number of cells requested at each refinement level.
            Defined in reversed order from the highest octree to lowest.
        :param finalize: Finalize the tree mesh after refinement.

        """
        if not isinstance(curve, Curve):
            raise TypeError("Refinement object must be a Curve.")

        if curve.vertices is None:
            return mesh

        if isinstance(levels, list):
            levels = np.array(levels)

        if isinstance(curve, BaseElectrode):
            locations = curve.vertices
        else:
            locations = densify_curve(curve, mesh.h[0][0])

        mesh = OctreeDriver.refine_tree_from_points(
            mesh, locations, levels, finalize=False
        )

        if finalize:
            mesh.finalize()

        return mesh

    @staticmethod
    def refine_tree_from_points(
        mesh: TreeMesh,
        points: ObjectBase | np.ndarray,
        levels: list[int] | np.ndarray,
        *,
        finalize: bool = False,
    ) -> TreeMesh:
        """
        Refine a tree mesh along the vertices of an object.

        :param mesh: Tree mesh to refine.
        :param points: Object to use for refinement.
        :param levels: Number of cells requested at each refinement level.
            Defined in reversed order from the highest octree to lowest.
        :param finalize: Finalize the tree mesh after refinement.

        :return: Refined tree mesh.
        """
        if isinstance(points, ObjectBase):
            locations = get_locations(points.workspace, points)
        else:
            locations = points

        if locations is None:
            raise ValueError("Could not find locations for refinement.")

        if isinstance(levels, list):
            levels = np.array(levels)

        distance = 0
        for ii, n_cells in enumerate(levels):
            distance += n_cells * OctreeDriver.cell_size_from_level(mesh, ii)
            mesh.refine_ball(
                locations,
                distance,
                mesh.max_level - ii,
                finalize=False,
            )

        if finalize:
            mesh.finalize()

        return mesh

    @staticmethod
    def refine_tree_from_surface(  # pylint: disable=too-many-locals
        mesh: TreeMesh,
        surface: ObjectBase,
        levels: list[int] | np.ndarray,
        *,
        max_distance: float = np.inf,
        finalize: bool = False,
    ) -> TreeMesh:
        """
        Refine a tree mesh along the simplicies of a surface.

        :param mesh: Tree mesh to refine.
        :param surface: Surface object to use for refinement.
        :param levels: Number of cells requested at each refinement level.
            Defined in reversed order from the highest octree to lowest.
        :param max_distance: Maximum distance from the surface to refine.
        :param finalize: Finalize the tree mesh after refinement.

        :return: Refined tree mesh.
        """
        if isinstance(levels, list):
            levels = np.array(levels)

        xyz = get_locations(surface.workspace, surface)
        triang = Delaunay(xyz[:, :2])

        tree = cKDTree(xyz[:, :2])
        interp = interpolate.LinearNDInterpolator(triang, xyz[:, -1])
        levels = np.array(levels)

        depth = 0
        # Cycle through the Tree levels backward
        for ind, n_cells in enumerate(levels):
            if n_cells == 0:
                continue

            dx = OctreeDriver.cell_size_from_level(mesh, ind, 0)
            dy = OctreeDriver.cell_size_from_level(mesh, ind, 1)
            dz = OctreeDriver.cell_size_from_level(mesh, ind, 2)

            # Create a grid at the octree level in xy
            assert surface.extent is not None
            cell_center_x, cell_center_y = np.meshgrid(
                np.arange(surface.extent[0, 0], surface.extent[1, 0], dx),
                np.arange(surface.extent[0, 1], surface.extent[1, 1], dy),
            )
            xy = np.c_[cell_center_x.reshape(-1), cell_center_y.reshape(-1)]

            # Only keep points within triangulation
            inside = triang.find_simplex(xy) != -1
            r, _ = tree.query(xy)
            keeper = np.logical_and(r < max_distance, inside)
            nnz = keeper.sum()
            elevation = interp(xy[keeper])

            # Apply vertical padding for current octree level
            for _ in range(int(n_cells)):
                depth += dz
                mesh.insert_cells(
                    np.c_[xy[keeper], elevation - depth],
                    np.ones(nnz) * mesh.max_level - ind,
                    finalize=False,
                )

        if finalize:
            mesh.finalize()

        return mesh

    @staticmethod
    def refine_tree_from_triangulation(
        mesh: TreeMesh,
        surface,
        levels: list[int] | np.ndarray,
        finalize=False,
    ) -> TreeMesh:
        """
        Refine a tree mesh along the simplicies of a surface.

        :param mesh: Tree mesh to refine.
        :param surface: Surface object to use for refinement.
        :param levels: Number of cells requested at each refinement level.
            Defined in reversed order from highest octree to lowest.
        :param finalize: Finalize the tree mesh after refinement.

        :return: Refined tree mesh.
        """
        if not isinstance(surface, Surface):
            raise TypeError("Refinement object must be a Surface.")

        if surface.vertices is None or surface.cells is None:
            raise ValueError("Surface object must have vertices and cells.")

        if isinstance(levels, list):
            levels = np.array(levels)

        vertices = surface.vertices.copy()
        normals = np.cross(
            vertices[surface.cells[:, 1], :] - vertices[surface.cells[:, 0], :],
            vertices[surface.cells[:, 2], :] - vertices[surface.cells[:, 0], :],
        )
        if surface.n_vertices is None:
            raise ValueError("Surface object must have n_vertices.")

        average_normals = np.zeros((surface.n_vertices, 3))

        for vert_ids in surface.cells.T:
            average_normals[vert_ids, :] += normals

        average_normals /= np.linalg.norm(average_normals, axis=1)[:, None]

        base_cells = np.r_[mesh.h[0][0], mesh.h[1][0], mesh.h[2][0]]
        for level, n_cells in enumerate(levels):
            if n_cells == 0:
                continue

            for _ in range(int(n_cells)):
                mesh.refine_surface(
                    (vertices, surface.cells),
                    level=-level - 1,
                    finalize=False,
                )
                vertices -= average_normals * base_cells * 2.0**level

        if finalize:
            mesh.finalize()

        return mesh

    @staticmethod
    def cell_size_from_level(octree, level: int, axis: int = 0):
        """
        Computes the cell size at a given level of refinement for a given tree mesh.

        :param octree: Tree mesh to refine.
        :param level: Level of refinement.
        :param axis: Axis of refinement.

        :return: Cell size at the given level of refinement.
        """
        return octree.h[axis][0] * 2**level

    @staticmethod
    def tree_offset(mesh: TreeMesh, vertices: np.ndarray) -> np.ndarray:
        """
        Compute the offset required to center the mesh around the vertices.

        :param mesh: Tree mesh to center.
        :param vertices: Vertices to center around.

        :return: Offset required to center
        """
        # Center on the nearest central vertices
        center = np.mean(vertices, axis=0)
        ind_mid = np.argmin(np.linalg.norm(vertices - center, axis=1))

        offsets = []
        for ii in range(mesh.dim):
            cell_centers = mesh.origin[ii] + np.cumsum(mesh.h[ii]) - mesh.h[ii] / 2
            nearest = np.searchsorted(cell_centers, vertices[ind_mid, ii])
            offsets.append(vertices[ind_mid, ii] - cell_centers[nearest])

        return np.r_[offsets]


if __name__ == "__main__":
    file = sys.argv[1]
    OctreeDriver.start(file)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2024-2025 Mira Geoscience Ltd.                                          '
#                                                                                        '
#  This file is part of octree-creation-app package.                                     '
#                                                                                        '
#  octree-creation-app is distributed under the terms and conditions of the MIT License  '
#  (see LICENSE file at the root of this source code package).                           '
#                                                                                        '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
from __future__ import annotations

from logging import warning

import discretize
import numpy as np
from discretize import TreeMesh
from geoh5py import Workspace
from geoh5py.objects import Curve, ObjectBase, Octree, Points
from geoh5py.shared.utils import fetch_active_workspace
from scipy.interpolate import interp1d
from scipy.spatial import cKDTree


def find_endpoints(points: np.ndarray) -> np.ndarray:
    """
    Find the endpoints of a co-linear array of points.

    :param points: locations array of shape (n, 3).
    """

    xmin = points[:, 0].min()
    xmax = points[:, 0].max()
    ymin = points[:, 1].min()
    ymax = points[:, 1].max()

    endpoints = []
    for x in np.unique([xmin, xmax]):
        for y in np.unique([ymin, ymax]):
            is_endy = np.isclose(points[:, :2], [x, y]).all(axis=1)
            if np.any(is_endy):
                endpoints.append(points[is_endy][0])

    return np.array(endpoints)


def surface_strip(
    points: ObjectBase, width: float, name: str = "Surface strip"
) -> Points:
    """
    Duplicate and offset co-linear input points to create a co-planar strip.

    :param points: Points object whose locations are all co-linear.
    :param width: Width used to displace existing points to create the
        strip.  The surrounding strip will be 2*width wider and longer than
        the input points.
    :param name: Name of the new Points objects.
    """

    assert points.locations is not None

    locs = points.locations
    ends = find_endpoints(locs)
    colinear = np.diff(ends[:, :2], axis=0)[0]
    colinear = colinear / np.linalg.norm(colinear)

    # Gram-Schmidt
    orthogonal = np.random.randn(2)
    orthogonal -= orthogonal.dot(colinear) * colinear
    orthogonal /= np.linalg.norm(orthogonal)

    vertices = np.vstack(
        [
            locs,
            np.r_[ends[1, :2] + width * colinear, ends[1, 2]],
            np.r_[ends[0, :2] - width * colinear, ends[0, 2]],
        ]
    )

    vertices = np.vstack(
        [
            vertices,
            np.c_[vertices[:, :2] + width * orthogonal, vertices[:, 2]],
            np.c_[vertices[:, :2] - width * orthogonal, vertices[:, 2]],
        ]
    )

    return Points.create(points.workspace, vertices=vertices, name=name)


def create_octree_from_octrees(meshes: list[Octree | TreeMesh]) -> TreeMesh:
    """
    Create an all encompassing octree mesh from a list of meshes.

    :param meshes: List of Octree or TreeMesh meshes.

    :return octree: A global Octree.
    """
    cell_size = []
    dimensions = None
    origin = None

    for mesh in meshes:
        attributes = get_octree_attributes(mesh)

        if dimensions is None:
            dimensions = attributes["dimensions"]
        elif not np.allclose(dimensions, attributes["dimensions"]):
            raise ValueError("Meshes must have same dimensions")

        if origin is None:
            origin = attributes["origin"]
        elif not np.allclose(origin, attributes["origin"]):
            raise ValueError("Meshes must have same origin")

        cell_size.append(attributes["cell_size"])

    cell_size = np.min(np.vstack(cell_size), axis=0)
    cells = []
    for ind in range(3):
        if dimensions is not None and cell_size is not None:
            extent = dimensions[ind]
            max_level = int(np.ceil(np.log2(np.abs(extent / cell_size[ind]))))
            cells += [np.ones(2**max_level) * cell_size[ind]]

    # Define the mesh and origin
    treemesh = TreeMesh(cells, origin=origin, diagonal_balance=False)

    for mesh in meshes:
        if isinstance(mesh, Octree) and mesh.octree_cells is not None:
            centers = mesh.centroids
            levels = treemesh.max_level - np.log2(mesh.octree_cells["NCells"])
        elif isinstance(mesh, TreeMesh) and mesh.cell_centers is not None:
            centers = mesh.cell_centers
            levels = (
                treemesh.max_level
                - mesh.max_level
                + mesh.cell_levels_by_index(np.arange(mesh.nC))
            )
        else:
            raise TypeError(
                f"All meshes must be Octree or TreeMesh, not {type(mesh)} "
                "and must have octree cells defined."
            )

        treemesh.insert_cells(centers, levels, finalize=False)

    treemesh.finalize()

    return treemesh


def collocate_octrees(global_mesh: Octree, local_meshes: list[Octree]):
    """
    Collocate a list of octree meshes into a global octree mesh.

    :param global_mesh: Global octree mesh.
    :param local_meshes: List of local octree meshes.
    """
    attributes = get_octree_attributes(global_mesh)
    cell_size = attributes["cell_size"]

    if (
        global_mesh.octree_cells is None
        or global_mesh.u_cell_size is None
        or global_mesh.v_cell_size is None
        or global_mesh.w_cell_size is None
    ):
        raise ValueError("Global mesh must have octree_cells and cell sizes.")

    u_grid = global_mesh.octree_cells["I"] * global_mesh.u_cell_size
    v_grid = global_mesh.octree_cells["J"] * global_mesh.v_cell_size
    w_grid = global_mesh.octree_cells["K"] * global_mesh.w_cell_size

    xyz = np.c_[u_grid, v_grid, w_grid] + attributes["origin"]
    tree = cKDTree(xyz)

    for local_mesh in local_meshes:
        attributes = get_octree_attributes(local_mesh)

        if cell_size and cell_size != attributes["cell_size"]:
            raise ValueError(
                f"Cell size mismatch in dimension {cell_size} != {attributes['cell_size']}"
            )

        _, closest = tree.query(attributes["origin"])
        shift = xyz[closest, :] - attributes["origin"]

        if np.any(shift != 0.0):
            with fetch_active_workspace(local_mesh.workspace) as workspace:
                warning(
                    f"Shifting {local_mesh.name} mesh origin by {shift} m to match inversion mesh."
                )
                local_mesh.origin = attributes["origin"] + shift
                workspace.update_attribute(local_mesh, "attributes")


def densify_curve(curve: Curve, increment: float) -> np.ndarray:
    """
    Refine a curve by adding points along the curve at a given increment.

    :param curve: Curve object to be refined.
    :param increment: Distance between points along the curve.

    :return: Array of shape (n, 3) of x, y, z locations.
    """
    locations = []
    for part in curve.unique_parts:
        if curve.cells is None or curve.vertices is None:
            continue

        logic = curve.parts == part
        cells = curve.cells[np.all(logic[curve.cells], axis=1)]

        if len(cells) == 0:
            continue

        vert_ind = np.r_[cells[:, 0], cells[-1, 1]]
        locs = curve.vertices[vert_ind, :]
        locations.append(resample_locations(locs, increment))

    if len(locations) == 0:
        return np.empty((0, 3))

    return np.vstack(locations)


def get_neighbouring_cells(mesh: TreeMesh, indices: list | np.ndarray) -> tuple:
    """
    Get the indices of neighbouring cells along a given axis for a given list of
    cell indices.

    :param mesh: discretize.TreeMesh object.
    :param indices: List of cell indices.

    :return: Two lists of neighbouring cell indices for every axis.
        axis[0] = (west, east)
        axis[1] = (south, north)
        axis[2] = (down, up)
    """
    if not isinstance(indices, (list, np.ndarray)):
        raise TypeError("Input 'indices' must be a list or numpy.ndarray of indices.")

    if not isinstance(mesh, TreeMesh):
        raise TypeError("Input 'mesh' must be a discretize.TreeMesh object.")

    neighbors: dict[int, list] = {ax: [[], []] for ax in range(mesh.dim)}

    for ind in indices:
        for ax in range(mesh.dim):
            neighbors[ax][0].append(np.r_[mesh[ind].neighbors[ax * 2]])
            neighbors[ax][1].append(np.r_[mesh[ind].neighbors[ax * 2 + 1]])

    return tuple(
        (np.r_[tuple(neighbors[ax][0])], np.r_[tuple(neighbors[ax][1])])
        for ax in range(mesh.dim)
    )


def get_octree_attributes(mesh: Octree | TreeMesh) -> dict[str, list]:
    """
    Get mesh attributes.

    :param mesh: Input Octree or TreeMesh object.
    :return mesh_attributes: Dictionary of mesh attributes.
    """
    if not isinstance(mesh, (Octree, TreeMesh)):
        raise TypeError(f"All meshes must be Octree or TreeMesh, not {type(mesh)}")

    cell_size = []
    cell_count = []
    dimensions = []
    if isinstance(mesh, TreeMesh):
        for int_dim in range(3):
            cell_size.append(mesh.h[int_dim][0])
            cell_count.append(mesh.h[int_dim].size)
            dimensions.append(mesh.h[int_dim].sum())
        origin = mesh.origin
    else:
        with fetch_active_workspace(mesh.workspace):
            for str_dim in "uvw":
                cell_size.append(getattr(mesh, f"{str_dim}_cell_size"))
                cell_count.append(getattr(mesh, f"{str_dim}_count"))
                dimensions.append(
                    getattr(mesh, f"{str_dim}_cell_size")
                    * getattr(mesh, f"{str_dim}_count")
                )
            origin = np.r_[mesh.origin["x"], mesh.origin["y"], mesh.origin["z"]]

    extent = np.r_[origin, origin + np.r_[dimensions]]

    return {
        "cell_count": cell_count,
        "cell_size": cell_size,
        "dimensions": dimensions,
        "extent": extent,
        "origin": origin,
    }


def octree_2_treemesh(  # pylint: disable=too-many-locals
    mesh: Octree,
) -> discretize.TreeMesh | None:
    """
    Convert a geoh5 octree mesh to discretize.TreeMesh

    Modified code from module discretize.TreeMesh.readUBC function.

    :param mesh: Octree mesh to convert.

    :return: Resulting TreeMesh.
    """
    if (
        mesh.octree_cells is None
        or mesh.u_count is None
        or mesh.v_count is None
        or mesh.w_count is None
    ):
        return None

    n_cell_dim, cell_sizes = [], []
    for ax in "uvw":
        if (
            getattr(mesh, f"{ax}_cell_size") is None
            or getattr(mesh, f"{ax}_count") is None
        ):
            raise ValueError(f"Cell size in {ax} direction is not defined.")

        n_cell_dim.append(getattr(mesh, f"{ax}_count"))
        cell_sizes.append(
            np.ones(getattr(mesh, f"{ax}_count")) * getattr(mesh, f"{ax}_cell_size")
        )

    if any(np.any(cell_size < 0) for cell_size in cell_sizes):
        raise NotImplementedError("Negative cell sizes not supported.")

    ls = np.log2(n_cell_dim).astype(int)

    if len(set(ls)) == 1:
        max_level = ls[0]
    else:
        max_level = min(ls) + 1

    cells = np.vstack(mesh.octree_cells.tolist())
    indexes = cells[:, :-1] * 2 + cells[:, -1][:, None]  # convert to cpp index
    levels = max_level - np.log2(cells[:, -1])
    treemesh = TreeMesh(
        cell_sizes, x0=np.asarray(mesh.origin.tolist()), diagonal_balance=False
    )
    treemesh.__setstate__((indexes, levels))

    return treemesh


def resample_locations(locations: np.ndarray, increment: float) -> np.ndarray:
    """
    Resample locations along a sequence of positions at a given increment.

    :param locations: Array of shape (n, 3) of x, y, z locations.
    :param increment: Minimum distance between points along the curve.

    :return: Array of shape (n, 3) of x, y, z locations.
    """
    distance = np.cumsum(
        np.r_[0, np.linalg.norm(locations[1:, :] - locations[:-1, :], axis=1)]
    )
    new_distances = np.sort(
        np.unique(np.r_[distance, np.arange(0, distance[-1], increment)])
    )

    resampled = []
    for axis in locations.T:
        interpolator = interp1d(distance, axis, kind="linear")
        resampled.append(interpolator(new_distances))

    return np.c_[resampled].T


def treemesh_2_octree(
    workspace: Workspace, treemesh: discretize.TreeMesh, **kwargs
) -> Octree:
    """
    Converts a :obj:`discretize.TreeMesh` to :obj:`geoh5py.objects.Octree` entity.

    :param workspace: Workspace to create the octree in.
    :param treemesh: TreeMesh to convert.

    :return: Octree entity.
    """

    if any(np.any(cell_size < 0) for cell_size in treemesh.h):
        raise NotImplementedError("Negative cell sizes not supported.")

    index_array = np.asarray(treemesh.cell_state["indexes"])
    levels = np.asarray(treemesh.cell_state["levels"])

    new_levels = 2 ** (treemesh.max_level - levels)
    new_index_array = (index_array - new_levels[:, None]) / 2

    origin = treemesh.x0.copy()

    mesh_object = Octree.create(
        workspace,
        origin=origin,
        u_count=treemesh.h[0].size,
        v_count=treemesh.h[1].size,
        w_count=treemesh.h[2].size,
        u_cell_size=treemesh.h[0][0],
        v_cell_size=treemesh.h[1][0],
        w_cell_size=treemesh.h[2][0],
        octree_cells=np.c_[new_index_array, new_levels],
        **kwargs,
    )

    return mesh_object
